{#
We want to extend the default template instead of defining everything ourselves.
#}
{% extends "default/module.html.jinja2" %}
{% macro is_public(doc) %}
    {% if doc.obj.__module__ is defined and doc.modulename != doc.obj.__module__ %}
    {% else %}
        {{ default_is_public(doc) }}
    {% endif %}
{% endmacro %}
